@@ -99,16 +99,17 @@ struct GSListenerPrivate {
 #endif
 
     guint32         ck_throttle_cookie;
+    int             sleep_inhibitor;
 };
 
 typedef struct {
-    int       entry_type;
-    char     *application;
-    char     *reason;
-    char     *connection;
-    guint32   cookie;
-    gint32    fd;
-    GTimeVal  since;
+    int        entry_type;
+    char      *application;
+    char      *reason;
+    char      *connection;
+    guint32    cookie;
+    gint32     fd;
+    GDateTime *since;
 } GSListenerRefEntry;
 
 enum {
@@ -160,6 +161,7 @@ gs_listener_ref_entry_free (GSListenerRefEntry *entry) {
     g_free (entry->connection);
     g_free (entry->application);
     g_free (entry->reason);
+    g_date_time_unref (entry->since);
     g_free (entry);
 }
 
@@ -311,6 +313,10 @@ listener_check_activation (GSListener *listener) {
         return TRUE;
     }
 
+    if (!listener->priv->session_idle) {
+        return TRUE;
+    }
+
     /* if we aren't inhibited then activate */
     inhibited = listener_ref_entry_is_present (listener, REF_ENTRY_TYPE_INHIBIT);
 
@@ -376,8 +382,6 @@ gs_listener_activate_saver (GSListener *listener,
     g_return_val_if_fail (GS_IS_LISTENER (listener), FALSE);
 
     if (listener->priv->active == active) {
-        gs_debug ("Trying to set active state when already: %s",
-                  active ? "active" : "inactive");
         return FALSE;
     }
 
@@ -387,6 +391,18 @@ gs_listener_activate_saver (GSListener *listener,
     return TRUE;
 }
 
+gboolean
+gs_listener_is_inhibited (GSListener *listener)
+{
+    gboolean inhibited;
+
+    g_return_val_if_fail (GS_IS_LISTENER (listener), FALSE);
+
+    inhibited = listener_ref_entry_is_present (listener, REF_ENTRY_TYPE_INHIBIT);
+
+    return inhibited;
+}
+
 static dbus_bool_t
 listener_property_set_bool (GSListener  *listener,
                             guint        prop_id,
@@ -482,6 +498,76 @@ listener_ref_entry_check (GSListener *listener,
     }
 }
 
+static void
+add_sleep_inhibit (GSListener *listener) {
+    DBusMessage     *message = NULL, *reply = NULL;
+    DBusMessageIter  iter, reply_iter;
+    DBusError        error;
+    const gchar     *what = "sleep";
+    const gchar     *who = "xfce4-screensaver";
+    const gchar     *why = "Locking screen before sleep";
+    const gchar     *mode = "delay";
+
+    g_return_if_fail (listener != NULL);
+
+    dbus_error_init (&error);
+
+#if defined(WITH_SYSTEMD) || defined(WITH_ELOGIND)
+    message = dbus_message_new_method_call (LOGIND_SERVICE,
+                                            LOGIND_PATH,
+                                            LOGIND_INTERFACE,
+                                            "Inhibit");
+#elif defined(WITH_CONSOLE_KIT)
+    message = dbus_message_new_method_call (CK_NAME,
+                                            CK_MANAGER_PATH,
+                                            CK_MANAGER_INTERFACE,
+                                            "Inhibit");
+#endif
+    if (message == NULL) {
+        gs_debug ("Couldn't allocate the dbus message");
+        return;
+    }
+
+    dbus_message_iter_init_append (message, &iter);
+    dbus_message_iter_append_basic (&iter, DBUS_TYPE_STRING, &what);
+    dbus_message_iter_append_basic (&iter, DBUS_TYPE_STRING, &who);
+    dbus_message_iter_append_basic (&iter, DBUS_TYPE_STRING, &why);
+    dbus_message_iter_append_basic (&iter, DBUS_TYPE_STRING, &mode);
+
+    reply = dbus_connection_send_with_reply_and_block (listener->priv->system_connection,
+                                                       message,
+                                                       -1,
+                                                       &error);
+    dbus_message_unref (message);
+
+    if (dbus_error_is_set (&error)) {
+        gs_debug ("%s raised:\n %s\n\n", error.name, error.message);
+        dbus_error_free (&error);
+        return;
+    }
+
+    dbus_message_iter_init (reply, &reply_iter);
+    if (DBUS_TYPE_UNIX_FD == dbus_message_iter_get_arg_type(&reply_iter))
+        dbus_message_iter_get_basic (&reply_iter, &listener->priv->sleep_inhibitor);
+
+    dbus_message_unref (reply);
+}
+
+static void
+remove_sleep_inhibit (GSListener *listener) {
+    g_return_if_fail (listener != NULL);
+
+    if (listener->priv->sleep_inhibitor < 0) {
+        gs_debug ("Can't remove sleep inhibitor: Lock not acquired");
+        return;
+    }
+
+    if (close(listener->priv->sleep_inhibitor) < 0) {
+        gs_debug ("Can't close file descriptor");
+        return;
+    }
+}
+
 static void
 add_session_inhibit (GSListener         *listener,
                      GSListenerRefEntry *entry) {
@@ -525,7 +611,7 @@ add_session_inhibit (GSListener         *listener,
     dbus_message_iter_append_basic (&iter, DBUS_TYPE_STRING, &mode);
 
     /* FIXME: use async? */
-    reply = dbus_connection_send_with_reply_and_block (listener->priv->connection,
+    reply = dbus_connection_send_with_reply_and_block (listener->priv->system_connection,
                                                        message,
                                                        -1,
                                                        &error);
@@ -617,6 +703,42 @@ out:
     return removed;
 }
 
+/**
+ * Borrowed from https://gitlab.gnome.org/GNOME/glib/blob/master/glib/gdatetime.c
+ * g_date_time_format_iso8601 is not available until 2.62, but we support 2.50+
+ * LGPL 2.1+
+ */
+static gchar *
+listener_g_date_time_format_iso8601 (GDateTime *datetime)
+{
+  GString *outstr = NULL;
+  gchar *main_date = NULL;
+  gint64 offset;
+
+  /* Main date and time. */
+  main_date = g_date_time_format (datetime, "%Y-%m-%dT%H:%M:%S");
+  outstr = g_string_new (main_date);
+  g_free (main_date);
+
+  /* Timezone. Format it as `%:::z` unless the offset is zero, in which case
+   * we can simply use `Z`. */
+  offset = g_date_time_get_utc_offset (datetime);
+
+  if (offset == 0)
+    {
+      g_string_append_c (outstr, 'Z');
+    }
+  else
+    {
+      gchar *time_zone = g_date_time_format (datetime, "%:::z");
+      g_string_append (outstr, time_zone);
+      g_free (time_zone);
+    }
+
+  return g_string_free (outstr, FALSE);
+}
+
+
 static void
 accumulate_ref_entry (gpointer            key,
                       GSListenerRefEntry *entry,
@@ -624,7 +746,7 @@ accumulate_ref_entry (gpointer            key,
     char *description;
     char *time;
 
-    time = g_time_val_to_iso8601 (&entry->since);
+    time = listener_g_date_time_format_iso8601 (entry->since);
 
     description = g_strdup_printf ("Application=\"%s\"; Since=\"%s\"; Reason=\"%s\";",
                                    entry->application,
@@ -692,7 +814,7 @@ listener_add_ck_ref_entry (GSListener     *listener,
     entry->cookie = listener_generate_unique_key (listener, entry_type);
     entry->application = g_strdup ("ConsoleKit");
     entry->reason = g_strdup ("Session is not active");
-    g_get_current_time (&entry->since);
+    entry->since = g_date_time_new_now_local ();
 
     /* takes ownership of entry */
     listener_add_ref_entry (listener, entry_type, entry);
@@ -749,7 +871,7 @@ listener_dbus_add_ref_entry (GSListener     *listener,
     entry->cookie = listener_generate_unique_key (listener, entry_type);
     entry->application = g_strdup (application);
     entry->reason = g_strdup (reason);
-    g_get_current_time (&entry->since);
+    entry->since = g_date_time_new_now_local ();
 
     listener_add_ref_entry (listener, entry_type, entry);
 
@@ -1366,10 +1488,16 @@ listener_dbus_handle_system_message (DBusConnection *connection,
                 if (listener->priv->prefs->sleep_activation_enabled) {
                     gs_debug ("Logind requested session lock");
                     g_signal_emit (listener, signals[LOCK], 0);
+
+                    gs_debug ("Releasing sleep inhibitor");
+                    remove_sleep_inhibit (listener);
                 } else {
                     gs_debug ("Logind requested session lock, but lock on suspend is disabled");
                 }
             } else {
+                gs_debug ("Reinstating logind sleep inhibitor lock");
+                add_sleep_inhibit (listener);
+
                 gs_debug ("Logind requested session unlock");
                 // FIXME: there is no signal to request password prompt
                 g_signal_emit (listener, signals[SHOW_MESSAGE], 0, NULL, NULL, NULL);
@@ -2061,6 +2189,9 @@ gs_listener_init (GSListener *listener) {
                                                         g_int_equal,
                                                         NULL,
                                                         (GDestroyNotify)gs_listener_ref_entry_free);
+
+    gs_debug ("Acquiring logind sleep inhibitor lock");
+    add_sleep_inhibit (listener);
 }
 
 static void
@@ -2083,6 +2214,7 @@ gs_listener_finalize (GObject *object) {
     }
 
     g_free (listener->priv->session_id);
+    remove_sleep_inhibit (listener);
 
     G_OBJECT_CLASS (gs_listener_parent_class)->finalize (object);
 }
